#include <pcap.h>
#include <assert.h>
#include <netinet/ether.h>
#include <netinet/ip.h>
#include <stdlib.h>
#include <netinet/in.h>

static struct entry {
  unsigned char mac[6];
  u_int32_t ip;
  struct entry * next;
} * top = NULL;

static int found_ip(u_int32_t ip) {
  struct entry * t;
  for (t=top;t!=NULL;t=t->next) {
    if (t->ip == ip)
      return 1;
  }
  return 0;
}

static void add_ip(unsigned char * mac, u_int32_t ip) {
  struct entry * t = malloc(sizeof(struct entry));
  assert(t!=NULL);
  memcpy(t->mac,mac,sizeof(t->mac));
  t->ip = ip;
  t->next = top;
  top = t;
}

static int is_ip(const u_char * packet) {
  struct ethhdr * ehdr = (struct ethhdr *)packet;
  if (ntohs(ehdr->h_proto) == ETH_P_IP) {
    return 1;
  }
  return 0;
}

static void extract_addr(const u_char * p, unsigned char * mac, u_int32_t * ip) {
  struct ethhdr * ehdr = (struct ethhdr *)p;
  struct iphdr * ihdr = (struct iphdr *)(p+sizeof(struct ethhdr));
  memcpy(mac,ehdr->h_source,sizeof(ehdr->h_source));
  *ip = ihdr->saddr;
} 

static int mac_cmp(unsigned char * mac1, unsigned char * mac2) {
  int i;
  for (i=0;i<6;i++) {
    if (mac1[i]!=mac2[i])
      return 0;
  }
  return 1;
}

static unsigned char * get_mac_from_ip(u_int32_t ip) {
  struct entry * t;
  for (t=top;t!=NULL;t=t->next) {
    if (t->ip == ip)
      return t->mac;
  }
  return NULL; /* XXX should not happen */
}

static inline void print_mac(unsigned char * mac) {
  printf("%02X:%02X:%02X:%02X:%02X:%02X\n",mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
}

static void print_clash(u_int32_t ip, unsigned char * mac1, unsigned char * mac2) {
  union ip_decode { u_int32_t ip; unsigned char nb[4]; } ipd;
  ipd.ip = ntohl(ip);
  printf("WARNING: found IP %d.%d.%d.%d with two different MAC addresses:\n",
         ipd.nb[3],ipd.nb[2],ipd.nb[1],ipd.nb[0]);
  print_mac(mac1);
  print_mac(mac2);
  puts("END WARNING");
}

int main(void) {
  char errbuf[PCAP_ERRBUF_SIZE];
  pcap_t * pcap_handle;
  const u_char * packet;
  struct pcap_pkthdr hdr;
  pcap_handle = pcap_open_live("eth0",1500,1,0,errbuf);
  assert(pcap_handle!=NULL);
  while (1) {
    packet = pcap_next(pcap_handle,&hdr);
    if (packet && is_ip(packet)) {
      unsigned char mac[6], * xmac;
      u_int32_t ip;
      extract_addr(packet,mac,&ip);
      if (!found_ip(ip)) {
        add_ip(mac,ip);
      } else if (mac_cmp((xmac=get_mac_from_ip(ip)),mac)==0) {
        print_clash(ip,mac,xmac);
      }
    }
  }
  return 0;
}
